"""
This module contains the synctranslations command for the Translations app.
"""

import sys

from django.core.management.base import (
    BaseCommand, CommandError,
)
from django.apps import apps
from django.db.models import Q
from django.contrib.contenttypes.models import ContentType

from translations.models import Translation, Translatable


__docformat__ = 'restructuredtext'


class Command(BaseCommand):
    """
    The command which synchronizes the translations with
    the apps models configurations.
    """

    help = 'Synchronize the translations with the apps models configurations.'

    def execute(self, *args, **options):
        """Execute the `Command` with `BaseCommand` arguments."""
        self.stdin = options.get('stdin', sys.stdin)  # Used for testing
        return super(Command, self).execute(*args, **options)

    def add_arguments(self, parser):
        """
        Add the arguments that the `Command` accepts on an `ArgumentParser`.
        """
        parser.add_argument(
            'args',
            metavar='app_label',
            nargs='*',
            help=(
                'Specify the app label(s) to synchronize the translations for.'
            ),
        )
        parser.add_argument(
            '--noinput', '--no-input',
            action='store_false',
            dest='interactive',
            help='Tells Django to NOT prompt the user for input of any kind.',
        )

    def get_content_types(self, *app_labels):
        r"""Return the `ContentType`\ s in some apps or all of them."""
        if app_labels:
            query = Q()
            for app_label in app_labels:
                try:
                    apps.get_app_config(app_label)
                except LookupError:
                    raise CommandError(
                        "App '{}' is not found.".format(app_label)
                    )
                else:
                    query |= Q(app_label=app_label)
            content_types = ContentType.objects.filter(query)
        else:
            content_types = ContentType.objects.all()
        return content_types

    def get_obsolete_translations(self, content_types):
        r"""Return the obsolete translations of some `ContentType`\ s."""
        if content_types:
            query = Q()
            for content_type in content_types:
                model = content_type.model_class()
                if issubclass(model, Translatable):
                    trans_fields = model._get_translatable_fields_names()
                    model_query = (
                        Q(content_type=content_type)
                        &
                        ~Q(field__in=trans_fields)
                    )
                else:
                    model_query = Q(content_type=content_type)
                query |= model_query
            obsolete_translations = Translation.objects.filter(query)
        else:
            obsolete_translations = Translation.objects.none()
        return obsolete_translations

    def log_obsolete_translations(self, obsolete_translations):
        """Log the details of some obsolete translations."""
        if self.verbosity >= 1:
            self.stdout.write('Looking for obsolete translations...')

            if obsolete_translations:
                changes = {}
                for translation in obsolete_translations:
                    app = apps.get_app_config(
                        translation.content_type.app_label
                    )
                    app_name = app.name
                    model = translation.content_type.model_class()
                    model_name = model.__name__

                    changes.setdefault(app_name, {})
                    changes[app_name].setdefault(model_name, set())
                    changes[app_name][model_name].add(translation.field)

                self.stdout.write(
                    'Obsolete translations found for the specified fields:'
                )

                for app_name, models in sorted(
                        changes.items(),
                        key=lambda x: x[0]):
                    self.stdout.write('- App: {}'.format(app_name))
                    for model_name, fields in sorted(
                            models.items(),
                            key=lambda x: x[0]):
                        self.stdout.write('  - Model: {}'.format(model_name))
                        for field in sorted(
                                fields,
                                key=lambda x: x[0]):
                            self.stdout.write('    - Field: {}'.format(field))

                self.stdout.write(
                    self.style.WARNING(
                        'Obsolete translations will be deleted in the '
                        'synchronization process.'
                    )
                )
            else:
                self.stdout.write('No obsolete translations found.')

    def ask_yes_no(self, message, default=None):
        """Ask the user for yes or no with a message and a default value."""
        answer = None
        while answer is None:
            value = input(message)

            # default
            if default is not None and value == '':
                value = default

            # yes or no?
            value = value.lower()
            if value in ['y', 'yes', True]:
                answer = True
            elif value in ['n', 'no', False]:
                answer = False
            else:
                answer = None
        return answer

    def should_run_synchronization(self):
        """Return whether to run the synchronization or not."""
        run = None
        if self.interactive:
            if hasattr(self.stdin, 'isatty') and not self.stdin.isatty():
                self.stderr.write(
                    "Synchronization failed due to not running in a TTY."
                )
                self.stderr.write(
                    "If you are sure about synchronization you can run "
                    "it with the '--no-input' flag."
                )
                sys.exit(1)
            else:
                try:
                    run = self.ask_yes_no(
                        (
                            'Are you sure you want to synchronize the '
                            'translations? [Y/n] '
                        ),
                        default='Y'
                    )
                except KeyboardInterrupt:
                    self.stdout.write('\n')  # move to the next line of stdin
                    self.stdout.write('\n')  # move another line for division
                    self.stderr.write("Operation cancelled.")
                    sys.exit(1)
        else:
            run = True

        return run

    def handle(self, *app_labels, **options):
        """Run the `Command` with the configured arguments."""
        # get arguments
        self.verbosity = options['verbosity']
        self.interactive = options['interactive']

        # collect all the models which will be affected
        content_types = self.get_content_types(*app_labels)

        # handle obsolete translations
        obsolete_translations = self.get_obsolete_translations(content_types)
        self.log_obsolete_translations(obsolete_translations)

        # divide initializing synchronization with asking for synchronization
        self.stdout.write('\n')

        if obsolete_translations:
            # ask user if they are sure that they want to synchronize
            run_synchronization = self.should_run_synchronization()

            # divide asking for synchronization with actual synchronization
            self.stdout.write('\n')

            if run_synchronization:
                obsolete_translations.delete()
            else:
                self.stdout.write(
                    'Synchronization cancelled.'
                )
                return

        self.stdout.write(
            self.style.SUCCESS(
                'Synchronization successful.'
            )
        )



"""This module contains the utilities for the Translations app."""

from django.db import models
from django.db.models.query import prefetch_related_objects
from django.db.models.constants import LOOKUP_SEP
from django.core.exceptions import FieldError
from django.contrib.contenttypes.models import ContentType
from django.utils.functional import SimpleLazyObject

import translations.models


__docformat__ = 'restructuredtext'


def _get_reverse_relation(model, relation):
    """Return the reverse of a model's relation."""
    parts = relation.split(LOOKUP_SEP)
    root = parts[0]
    branch = parts[1:]

    field = model._meta.get_field(root)
    reverse_relation = field.remote_field.name

    if branch:
        branch_model = field.related_model
        branch_relation = LOOKUP_SEP.join(branch)
        branch_reverse_relation = _get_reverse_relation(
            branch_model,
            branch_relation
        )
        return '{}__{}'.format(
            branch_reverse_relation,
            reverse_relation,
        )
    else:
        return reverse_relation


def _get_dissected_lookup(model, lookup):
    """Return the dissected info of a lookup."""
    dissected = {
        'relation': [],
        'field': '',
        'supplement': '',
        'translatable': False,
    }

    def _fill_dissected(model, *relation_parts):
        root = relation_parts[0]
        nest = relation_parts[1:]

        try:
            if root == 'pk':
                field = model._meta.pk
            else:
                field = model._meta.get_field(root)
        except Exception as e:
            if not dissected['relation'] or nest or dissected['field']:
                raise e
            dissected['supplement'] = root
        else:
            field_model = field.related_model
            if field_model:
                dissected['relation'].append(root)
                if nest:
                    _fill_dissected(field_model, *nest)
            else:
                dissected['field'] = root
                if issubclass(model, translations.models.Translatable):
                    if root in model._get_translatable_fields_names():
                        dissected['translatable'] = True
                if nest:
                    if len(nest) == 1:
                        dissected['supplement'] = nest[0]
                    else:
                        raise FieldError("Unsupported lookup '{}'".format(
                            nest[0])
                        )

    parts = lookup.split(LOOKUP_SEP)

    _fill_dissected(model, *parts)

    return dissected


def _get_relations_hierarchy(*relations):
    """Return the relations hierarchy of some relations."""
    hierarchy = {}

    def _fill_hierarchy(hierarchy, *relation_parts):
        root = relation_parts[0]
        nest = relation_parts[1:]

        hierarchy.setdefault(root, {
            'included': False,
            'relations': {},
        })

        if nest:
            _fill_hierarchy(hierarchy[root]['relations'], *nest)
        else:
            hierarchy[root]['included'] = True

    for relation in relations:
        parts = relation.split(LOOKUP_SEP)
        _fill_hierarchy(hierarchy, *parts)

    return hierarchy


def _get_entity_details(entity):
    """Return the iteration and type details of an entity."""

    error_message = SimpleLazyObject(
        lambda: '`{}` is neither {} nor {}.'.format(
            entity,
            'a model instance',
            'an iterable of model instances',
        )
    )

    if isinstance(entity, models.Model):
        model = type(entity)
        iterable = False
    elif hasattr(entity, '__iter__'):
        if len(entity) > 0:
            if isinstance(entity[0], models.Model):
                model = type(entity[0])
            else:
                raise TypeError(error_message)
        else:
            model = None
        iterable = True
    else:
        raise TypeError(error_message)

    return (iterable, model)


def _get_purview(entity, hierarchy):
    """Return the purview of an entity and a relations hierarchy of it."""
    mapping = {}
    query = models.Q()

    def _fill_entity(entity, hierarchy, included=True):
        iterable, model = _get_entity_details(entity)

        if model is None:
            return

        content_type_id = ContentType.objects.get_for_model(model).id

        if included:
            instances = mapping.setdefault(content_type_id, {})
            if not issubclass(model, translations.models.Translatable):
                raise TypeError('`{}` is not Translatable!'.format(model))

        def _fill_obj(obj):
            if included:
                if not hasattr(obj, '_default_translatable_fields'):
                    obj._default_translatable_fields = {
                        field: getattr(obj, field) for field in
                        type(obj)._get_translatable_fields_names()
                    }
                object_id = str(obj.pk)
                instances[object_id] = obj
                nonlocal query
                query |= models.Q(
                    content_type__id=content_type_id,
                    object_id=object_id,
                )

            if hierarchy:
                for (relation, detail) in hierarchy.items():
                    value = getattr(obj, relation, None)

                    if value is not None:
                        if isinstance(value, models.Manager):
                            if not (
                                hasattr(obj, '_prefetched_objects_cache') and
                                relation in obj._prefetched_objects_cache
                            ):
                                prefetch_related_objects([obj], relation)
                            value = value.all()
                        _fill_entity(
                            entity=value,
                            hierarchy=detail['relations'],
                            included=detail['included'],
                        )

        if iterable:
            for obj in entity:
                _fill_obj(obj)
        else:
            _fill_obj(entity)

    _fill_entity(entity, hierarchy)

    return mapping, query


def _get_translations(query, lang):
    """Return the `Translation` queryset of a query in a language."""
    if (query):
        queryset = translations.models.Translation.objects.filter(
            language=lang,
        ).filter(
            query,
        ).select_related('content_type')

        return queryset
    else:
        return translations.models.Translation.objects.none()



"""This module contains the languages for the Translations app."""

from django.utils.translation import get_language
from django.conf import settings


__docformat__ = 'restructuredtext'


_supported_code = {}

_all_codes = None
_all_choices = None

_translation_codes = {}
_translation_choices = {}


def _get_supported_language(lang):
    """Return the `supported language` code of a custom language code."""
    if lang not in _supported_code:
        code = lang.split('-')[0]

        lang_exists = False
        code_exists = False

        # break when the lang is found but not when the code is found
        # cause the code might come before lang and we may miss an accent
        for choice in settings.LANGUAGES:
            if lang == choice[0]:
                lang_exists = True
                break
            if code == choice[0]:
                code_exists = True

        if lang_exists:
            _supported_code[lang] = lang
        elif code_exists:
            _supported_code[lang] = code
        else:
            raise ValueError(
                '`{}` is not a supported language.'.format(lang)
            )
    return _supported_code[lang]


def _get_default_language():
    """Return the `supported language` code of the `default language` code."""
    return _get_supported_language(settings.LANGUAGE_CODE)


def _get_active_language():
    """Return the `supported language` code of the `active language` code."""
    return _get_supported_language(get_language())


def _get_all_languages():
    """Return all the `supported language` codes."""
    global _all_codes
    if _all_codes is None:
        _all_codes = [choice[0] for choice in settings.LANGUAGES]
    return _all_codes


def _get_all_choices():
    """Return all the `supported language` choices."""
    global _all_choices
    if _all_choices is None:
        _all_choices = [
            (None, '---------'),
        ] + [choice for choice in settings.LANGUAGES]
    return _all_choices


def _get_translation_languages():
    """Return the `translation language` codes."""
    default = _get_default_language()
    if default not in _translation_codes:
        _translation_codes[default] = [
            lang for lang in _get_all_languages() if lang != default]
    return _translation_codes[default]


def _get_translation_choices():
    """Return the `translation language` choices."""
    default = _get_default_language()
    if default not in _translation_choices:
        _translation_choices[default] = [
            choice for choice in _get_all_choices() if choice[0] != default]
    return _translation_choices[default]


def _get_translate_language(lang=None):
    """Return the `supported language` code of a translate language code."""
    if lang is None:
        return _get_active_language()
    else:
        return _get_supported_language(lang)


def _get_probe_language(lang=None):
    """
    Return the `supported language` code(s) of some probe language code(s).
    """
    if isinstance(lang, (list, tuple)):
        return [_get_supported_language(x) for x in lang]
    else:
        return _get_translate_language(lang)


class _TRANSLATE:
    """A class which provides standard translate language codes."""

    @property
    def DEFAULT(self):
        """Return the `default language`."""
        return _get_default_language()

    @property
    def ACTIVE(self):
        """Return the `active language` code."""
        return _get_active_language()


class _PROBE:
    """A class which provides standard probe language codes."""

    @property
    def DEFAULT(self):
        """Return the `default language` code."""
        return _get_default_language()

    @property
    def ACTIVE(self):
        """Return the `active language` code."""
        return _get_active_language()

    @property
    def DEFAULT_ACTIVE(self):
        """Return the `default language` and `active language` codes."""
        if self.DEFAULT != self.ACTIVE:
            return [self.DEFAULT, self.ACTIVE]
        else:
            return self.DEFAULT

    @property
    def TRANSLATION(self):
        """Return the `translation language` codes."""
        return _get_translation_languages()

    @property
    def ALL(self):
        """Return all the `supported language` codes."""
        return _get_all_languages()


translate = _TRANSLATE()
"""An object which provides standard translate language codes."""

probe = _PROBE()
"""An object which provides standard probe language codes."""



from django.apps import AppConfig
try:
    from django.utils.translation import ugettext_lazy as _
except ImportError:
    from django.utils.translation import gettext_lazy as _


class TranslationsConfig(AppConfig):
    name = 'translations'
    verbose_name = _('translations')



import setuptools
import json


with open('./.github/README.md', 'r') as fh:
    long_description = fh.read()

with open('config.json', 'r') as fh:
    info = json.load(fh)


setuptools.setup(
    name=info['project']['dist'],
    version=info['release']['name'],
    description=info['project']['desc'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=info['urls']['source'],
    author=info['author']['name'],
    author_email=info['author']['email'],
    classifiers=[
        'Development Status :: ' + info['release']['classifier'],
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.1',
        'Framework :: Django :: 4.2',
        'Framework :: Django :: 5.0',
        'Framework :: Django :: 5.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Topic :: Software Development :: Internationalization',
    ],
    keywords=' '.join(info['keywords']),
    project_urls={
        'Documentation': info['urls']['documentation'],
        'Funding': info['urls']['funding'],
        'Source': info['urls']['source'],
        'Tracker': info['urls']['tracker'],
    },
    packages=setuptools.find_packages(
        exclude=[
            'sample',
            'sample.migrations',
            'tests',
        ]
    ),
    python_requires='>=3.7, <4',
)



"""This module contains the admins for the Translations app."""

from django.contrib.contenttypes.admin import GenericStackedInline
from django.contrib import admin

from translations.models import Translation
from translations.forms import generate_translation_form


__docformat__ = 'restructuredtext'


class TranslatableAdminMixin:
    """An admin mixin which provides custom translation functionalities."""

    def prepare_translation_inlines(self, inlines, inline_type):
        """
        Prepare the translation inlines of a type in some inlines based on the
        admin model.
        """
        form = generate_translation_form(self.model)
        remove_inlines = []
        for i, v in enumerate(inlines):
            if isinstance(v, inline_type):
                if len(form.base_fields['field'].choices) == 1:
                    remove_inlines.append(i)
                else:
                    inlines[i].form = form
        remove_inlines.reverse()
        for index in remove_inlines:
            inlines.pop(index)


class TranslatableAdmin(TranslatableAdminMixin, admin.ModelAdmin):
    """The admin which represents the `Translatable` instances."""

    def get_inline_instances(self, request, obj=None):
        inlines = list(
            super(TranslatableAdmin, self).get_inline_instances(request, obj)
        )
        self.prepare_translation_inlines(inlines, TranslationInline)
        return inlines


class TranslationInline(GenericStackedInline):
    """The inline which represents the `Translation` instances."""

    model = Translation
    extra = 1
